use std::ffi::OsStr;
use std::fs::File;
use std::io::prelude::*;
use std::path::PathBuf;
use udev::Device;

const CMEDIA_VID: u16 = 0xd8c;
const CMEDIA_PID1_MIN: u16 = 0x0008;
const CMEDIA_PID1_MAX: u16 = 0x000f;

const CMEDIA_PIDS: [u16; 5] = [0x0139, 0x0012, 0x013a, 0x0013, 0x013c];

#[derive(Debug)]
pub struct Thing {
	vendor_id: u16,
	product_id: u16,
	product: String,
	sound_devnode: Option<PathBuf>,
	usb_devnode: PathBuf,
	hidraw_devnode: Option<PathBuf>,
}

impl Thing {
	pub fn set_all_gpio(&self, on: bool) -> Option<()> {
		if on {
			self.write(255, 255)
		} else {
			self.write(255, 0)
		}
	}
	/*
		pub fn set_gpio(&self, num: u8, on: bool) -> Option<()> {
			if !(1..=8).contains(&num) {
				return None;
			}

			let iomask = 1 << (num - 1);
			let iodata = if on { 1 << (num - 1) } else { 0 };

			self.write(iomask, iodata)
		}
	*/
	fn write(&self, iomask: u8, iodata: u8) -> Option<()> {
		let mut f = File::create(self.hidraw_devnode.as_ref()?).ok()?;
		f.write_all(&[0, 0, iomask, iodata, 0]).ok()?;

		Some(())
	}
}

pub fn find_cm108_device() -> Option<Thing> {
	let mut things = Vec::new();
	let mut enumerator = udev::Enumerator::new().ok()?;
	enumerator.match_subsystem("sound").ok()?;
	for device in enumerator.scan_devices().ok()? {
		if let Some(t) = process_device(device) {
			things.push(t);
		}
	}

	let mut enumerator = udev::Enumerator::new().ok()?;
	enumerator.match_subsystem("hidraw").ok()?;
	for device in enumerator.scan_devices().ok()? {
		if let Some(mut t) = process_device(device) {
			let mut found = false;
			for old_t in &mut things {
				if old_t.vendor_id == t.vendor_id
					&& old_t.product_id == t.product_id
					&& old_t.usb_devnode == t.usb_devnode
				{
					found = true;
					old_t.hidraw_devnode = t.sound_devnode.clone();
				}
			}

			if !found {
				t.hidraw_devnode = t.sound_devnode;
				t.sound_devnode = None;
				things.push(t);
			}
		}
	}

	things.into_iter().find(|t| {
		t.vendor_id == CMEDIA_VID
			&& ((t.product_id >= CMEDIA_PID1_MIN && t.product_id <= CMEDIA_PID1_MAX)
				|| CMEDIA_PIDS.contains(&t.product_id))
	})
}

fn process_device(device: Device) -> Option<Thing> {
	let dev_node = device.devnode()?;
	let parent_dev = device
		.parent_with_subsystem_devtype("usb", "usb_device")
		.ok()??;
	let vendor_id = os2u16(parent_dev.attribute_value("idVendor")?)?;
	let product_id = os2u16(parent_dev.attribute_value("idProduct")?)?;
	let product = parent_dev.attribute_value("product")?.to_str()?.to_string();
	let sound_devnode = dev_node.to_owned();
	let usb_devnode = parent_dev.devnode()?.to_owned();

	Some(Thing {
		vendor_id,
		product_id,
		product,
		sound_devnode: Some(sound_devnode),
		usb_devnode,
		hidraw_devnode: None,
	})
}

fn os2u16<T: AsRef<OsStr>>(s: T) -> Option<u16> {
	u16::from_str_radix(s.as_ref().to_str()?, 16).ok()
}
